/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Matthew Wilson <matthew@mjwilson.demon.co.uk>.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*************************
 * requires:
 *   chrome://inspector/content/jsutil/xpcom/XPCU.js
 *   chrome://inspector/content/jsutil/rdf/RDFU.js
 */

function AnnozillaPrefs(annozilla) {
    var that = this;

    this.annozilla = annozilla;

    // get the AnnotationsApplication
    var annotService = Components.classes[AnnozillaPrefs.ANNOTATIONS_SERVICE_CONTRACTID].getService().QueryInterface(Components.interfaces.nsIAnnotationsService);
    this.app = annotService.registerApplication("annozilla");

    // set up the prefs themselves
    this.prefValues = {
        annotationServer:   undefined,  // URL of server, used up to 0.3 versions, only kept for migration
        annotationUser:     undefined,  // user name, used up to 0.3 versions, only kept for migration
        annotationPassword: undefined,  // user password, used up to 0.3 versions, only kept for migration
        defaultPostServer:  undefined,  // URL of post server, used from 0.4 onwards
        defaultPostServerSet: function() {
            that.setDefaultPostServer();
        },
        debug:              undefined,
        insertIcons:        undefined,
        insertIconsSet:     function() {
            if (that.annozilla.insertIconsPrefChanging) {
                that.annozilla.insertIconsPrefChanging(this.insertIcons);
            }
        },
        xbl:                undefined,
        autoLoadMulti:      undefined, // Whether to auto-load annotations: "never"/"always"/"sites"
        autoLoadMultiSet:   function() {
            if (this.autoLoadMulti != "never") {
                that.app.autoLoadAnnotations(true);
            }
            else {
                that.app.autoLoadAnnotations(false);
            }
        },
        autoLoadSites:      undefined,
        language:           undefined,
        showOnlyMyLanguage: undefined,
        showWhenNoLanguage: undefined,
        creator: undefined,
        keepEditWindowOpen:  undefined,
        loadInNewWindow: undefined
    }
    
    // set up the change observer
    this.changeObserver = {
        observe: function(aSubject, aTopic, aData)
        {
            var prefvar = aData.substring(AnnozillaPrefs.PREFIX.length); // strip off prefix
            that.prefValues[prefvar] = that.loadPref(aData);
            if (typeof that.prefValues[prefvar + "Set"] == "function") {
                that.prefValues[prefvar + "Set"] (that.prefValues[prefvar]);
            } 
        }
    };

    /* add change observer for preferences prefixed by "annozilla" */
    this.prefService.addObserver ("annozilla", this.changeObserver, false);

    for (pref in this.prefValues) {
        if (typeof this.prefValues[pref] != "function") {
            this.prefValues[pref] = this.loadPref(AnnozillaPrefs.PREFIX + pref);
            if (typeof this.prefValues[pref + "Set"] == "function") {
                this.prefValues[pref + "Set"] (this.prefValues[pref]);
            }
        }
    }

    // load the servers datasource
    this.serverDS = this.loadServersDS();

    // add the servers to the annotations application
    this.addServersToService();

    // add an observer to maintain these servers
    this.serverObserver = new ServerObserver(this.app, this);
    this.serverDS.AddObserver(this.serverObserver);        

    // set up a default post server
    this.setDefaultPostServer();
}

// CLASS FIELDS
AnnozillaPrefs.ANNOTATIONS_SERVICE_CONTRACTID = "@mozilla.org/annotations_service;1";
AnnozillaPrefs.DIRECTORY_SERVICE_CONTRACTID = "@mozilla.org/file/directory_service;1";
AnnozillaPrefs.NS_LOCAL_FILE_CONTRACTID = "@mozilla.org/file/local;1";
AnnozillaPrefs.NS_LOCALFILEOUTPUTSTREAM_CONTRACTID = "@mozilla.org/network/file-output-stream;1";

AnnozillaPrefs.PROFILE_DIRECTORY_ID = "ProfD";
AnnozillaPrefs.SERVERS_FILENAME = "annozillaServers.rdf";
AnnozillaPrefs.SERVERS_ROOT_URN = "urn:annotations:servers";

// File read/write operations
AnnozillaPrefs.NS_RDONLY      = 0x01;
AnnozillaPrefs.NS_WRONLY      = 0x02;
AnnozillaPrefs.NS_CREATE_FILE = 0x08;
AnnozillaPrefs.NS_TRUNCATE    = 0x20;

// user -- read, write; group -- read; other -- read

AnnozillaPrefs.DEFAULT_FILE_PERMS = 00400 | 00200 | 00040 | 00004;

AnnozillaPrefs.NS_PREF_INTERFACE = Components.interfaces.nsIPref;
AnnozillaPrefs.PREFIX = "annozilla.";
AnnozillaPrefs._defaultValues = {  loadInNewWindow: { value: false,
                                                      type:  AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool },
                                   keepEditWindowOpen:{ value: false,
                                                        type:  AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool },
                                   creator: { value: "",
                                              type: AnnozillaPrefs.NS_PREF_INTERFACE.ePrefString },
                                   insertIcons: { value: false,
                                                 type:  AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool },
                                  xbl:         { value: false,
                                                 type:  AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool },
                                  autoLoadMulti: { value: "never",
                                                   type: AnnozillaPrefs.NS_PREF_INTERFACE.ePrefString },
                                  language:  { value: "en",
                                               type: AnnozillaPrefs.NS_PREF_INTERFACE.ePrefString },
                                  showOnlyMyLanguage: { value: false,
                                                        type:  AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool },
                                  showWhenNoLanguage: { value: true,
                                                        type:  AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool } };

AnnozillaPrefs.prototype = {
    prefService:  Components.classes["@mozilla.org/preferences;1"].getService(Components.interfaces[Components.interfaces.nsIPref]),
    changeObserver: undefined,
    prefValues:    undefined
};

AnnozillaPrefs.prototype.destructor = function() {
    this.prefService.removeObserver("annozilla", this.changeObserver);
    this.serverDS.RemoveObserver(this.serverObserver);
}

AnnozillaPrefs.prototype.getPref = function(aPrefName) 
{
    return this.prefValues[aPrefName] ? this.prefValues[aPrefName] : null;
}

AnnozillaPrefs.prototype.loadPref = function(fullName)
{
    var type = this.prefService.GetPrefType(fullName);
    
    if (type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefString) {
        return this.prefService.CopyUnicharPref(fullName);
    } 
    else if (type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool) {
        return this.prefService.GetBoolPref(fullName);
    } 
    else if (type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefInt) {
        return this.prefService.GetIntPref(fullName);
    }
    else if (type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefInvalid) {
        var shortName = fullName.substring(AnnozillaPrefs.PREFIX.length); // strip off prefix
        if (AnnozillaPrefs._defaultValues[shortName]) {
            
            if (AnnozillaPrefs._defaultValues[shortName].type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefString) {
                this.prefService.SetUnicharPref(fullName, AnnozillaPrefs._defaultValues[shortName].value);
            } 
            else if (AnnozillaPrefs._defaultValues[shortName].type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefBool) {
                this.prefService.SetBoolPref(fullName, AnnozillaPrefs._defaultValues[shortName].value);
            } 
            else if (AnnozillaPrefs._defaultValues[shortName].type == AnnozillaPrefs.NS_PREF_INTERFACE.ePrefInt) {
                this.prefService.SetIntPref(fullName, AnnozillaPrefs._defaultValues[shortName].value);
            }
            else {
                throw new Error("Invalid default pref type in loadPref");
            }
            return AnnozillaPrefs._defaultValues[shortName].value;
        }
        else {
            return null;
        }
    }
    else {
        return null;
    }
}

AnnozillaPrefs.prototype.loadServersDS = function() {
    // get the preferences directory
    var fileLocator = Components.classes[AnnozillaPrefs.DIRECTORY_SERVICE_CONTRACTID].getService(Components.interfaces.nsIProperties);  
    var file = fileLocator.get("ProfD", Components.interfaces.nsIFile);
  
    // append servers filename
    file.append(AnnozillaPrefs.SERVERS_FILENAME);        
    
    // if the file does not exist, create it
    if (!file.exists()) {
        dump("creating new annozilla servers file\n");
        AnnozillaPrefs.createNewAnnozillaServersFile(file, this.prefValues);
    }                
                
    // load the data source
    var urlSpec = AnnozillaPrefs.getURLSpecFromFile(file);
    return gRDF.GetDataSourceBlocking(urlSpec);
}
        
/**
 * 
 * @param dataSource servers data source
 */
AnnozillaPrefs.prototype.addServersToService = function() {

    // build a seq container
    var seq = RDFU.makeSeq(this.serverDS, gRDF.GetResource(AnnozillaPrefs.SERVERS_ROOT_URN));

    // enumerate it and add these servers
    var enumeration = seq.GetElements();
    while (enumeration.hasMoreElements()) {
        // load the list element in the Bag
        var element = enumeration.getNext().QueryInterface(Components.interfaces.nsIRDFResource);

        // load the actual server resource
        var resource = gRDF.GetResource(element.Value).QueryInterface(Components.interfaces.nsIRDFResource);
        var name = resource.Value;
        var user = this.getLiteralFromResource(this.serverDS, resource, "user");
        var password = this.getLiteralFromResource(this.serverDS, resource, "password");

        // add the server -- if it already exists, this is fine.  it
        // will do nothing
        this.app.addServer(name, user, password);
    }
}
 
AnnozillaPrefs.prototype.setDefaultPostServer = function() {
    var defaultServer = this.getPref("defaultPostServer");

    if (defaultServer && this.app && this.app.hasServer(defaultServer)) {
        this.app.setDefaultPostServer(defaultServer);
    }
}


/**
 * Get a literal's String value from a resource in the global
 * datasource.
 *
 * @param aResource Resource to change the literal for
 * @param aPropertyString literal to get
 * @return the String value of that literal
 */
AnnozillaPrefs.prototype.getLiteralFromResource = function(aDataSource, aResource, aPropertyString) {
    try {
        var valueProperty = gRDF.GetResource(AnnozillaPrefs.SERVER_RDF(aPropertyString));
        var node = aDataSource.GetTarget(aResource, valueProperty, true);
        return node.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    }
    catch(e) {
        dump("AnnozillaPrefs:  getLiteralFromResource error: " + e + "\n");
        return null;
    }
}



/**
 * Creates a new servers file.
 * The file is populated with existing server data from 0.3 versions,
 * if that data exists; otherwise it is created blank.
 *
 * @param file nsIFile with location for new file
 * @param prefValues Optional preference object containing server, user, password
 * @return the nsIFile (evals to a nsILocalFile) created
 */
AnnozillaPrefs.createNewAnnozillaServersFile = function(file, prefValues) {
    try {
        
        // initialize local file
        var localFile = Components.classes[AnnozillaPrefs.NS_LOCAL_FILE_CONTRACTID].createInstance(Components.interfaces.nsILocalFile);
        
        localFile.initWithPath(file.path);

        if (localFile.exists() && 
            (localFile.isDirectory() || !localFile.isWritable())) {
            throw Components.results.NS_ERROR_FAILURE;                
        }
        
        // write flags
        var flags = AnnozillaPrefs.NS_WRONLY | AnnozillaPrefs.NS_CREATE_FILE | AnnozillaPrefs.NS_TRUNCATE;
        
        // get a file output stream
        var fileStream = Components.classes[AnnozillaPrefs.NS_LOCALFILEOUTPUTSTREAM_CONTRACTID].createInstance(Components.interfaces.nsIFileOutputStream);
        fileStream.init(localFile, flags, AnnozillaPrefs.DEFAULT_FILE_PERMS, 0);

        // template for servers rdf
        var outputStr = "<?xml version=\"1.0\"?>\n" +
            "<RDF:RDF xmlns:s=\"urn:annotations:serverNS#\"  \n" +
            "   xmlns:NC=\"http://home.netscape.com/NC-rdf#\"  \n" +
            "   xmlns:RDF=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n" +
            "   <RDF:Bag about=\"urn:annotations:servers\">\n" +
            "   </RDF:Bag>\n" +
            "</RDF:RDF>\n";

        // Migrate old (0.3) style preferences
        // TODO: Set defaultPostServer
        if (prefValues != undefined &&
            prefValues.annotationServer   != undefined &&
            prefValues.annotationUser     != undefined &&
            prefValues.annotationPassword != undefined) {
            outputStr =
                "<?xml version=\"1.0\"?>\n" +
                "<RDF:RDF xmlns:s=\"urn:annotations:serverNS#\"\n" +
                "         xmlns:NC=\"http://home.netscape.com/NC-rdf#\"\n" +
                "         xmlns:RDF=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n" +
                "  <RDF:Bag about=\"urn:annotations:servers\">\n" +
                "    <RDF:li resource=\"" + prefValues.annotationServer + "\"/>\n" +
                "  </RDF:Bag>\n" +
                "  <RDF:Description about=\"" + prefValues.annotationServer + "\"\n" +
                "                   s:user=\"" + prefValues.annotationUser + "\"\n" +
                "                   s:password=\"" + prefValues.annotationPassword + "\" />\n" +
                "</RDF:RDF>\n";
        }
        
        fileStream.write(outputStr, outputStr.length);
        fileStream.close();

        return localFile;
    }
    catch (e) {
        dump("Error writing out new file: " + e + "\n");
        return undefined;
    }
}
        

AnnozillaPrefs.getURLSpecFromFile = function(file) {
    
    if (Components.classes["@mozilla.org/network/io-service;1"] && Components.interfaces.nsIIOService) {
        
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
        var fileHandler = ioService.getProtocolHandler("file").QueryInterface(Components.interfaces.nsIFileProtocolHandler);

        // new version
        if (fileHandler.getURLSpecFromFile) {
            return fileHandler.getURLSpecFromFile(file);
        }
        // old version in ioService
        else {
            return ioService.getURLSpecFromFile(file);
        }
    }
    // really old version
    else {
        var file_url = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIFileURL);
        file_url.file = file;
        return file_url.spec;
    }
}
/**
 * Prepend the server namespace string to a property.
 *
 * @param aProperty String property
 * @return concatenated String
 */
AnnozillaPrefs.SERVER_RDF = function(aProperty)
{
    return "urn:annotations:serverNS#" + aProperty;
}

/**
 * Implementation of nsIRDFObserver.
 * Observes server rdf and updates servers
 * when necessary.
 *
 * @param prefs AnnozillaPrefs creating this
 */
function ServerObserver(app, prefs) { 
    this.app = app;
    this.prefs = prefs;
    this.newServers = {};
    this.changeServers = {};
}

// regExp for screening out the top level container of servers
ServerObserver.topLevelRegExp = new RegExp("^urn:annotations:servers");

ServerObserver.SERVER_NAMESPACE = "urn:annotations:serverNS#";  // the server ns

ServerObserver.prototype = {
    app:        undefined,
    newServers: undefined,
    changeServers: undefined,
    prefs:      undefined
};

/**
 * Interface recognition.
 *
 * @param iid nsISupports Interface queried
 * @return this if requesting nsIRDFObserver, error otherwise
 */
ServerObserver.prototype.QueryInterface = function (iid) {
    
    if (!iid.equals(Components.interfaces.nsIRDFObserver) && !iid.equals(Components.interfaces.nsISupports))
    throw Components.results.NS_ERROR_NO_INTERFACE;
    
    return this;
}

/**
 * This method is called whenever a new assertion is made
 * in the data source.  A new assertion == a new server.
 *
 * @param aDataSource nsIRDFDataSource the datasource that is issuing
 *   the notification.
 * @param aSource nsIRDFResource the subject of the assertion
 * @param aProperty nsIRDFResource the predicate of the assertion
 * @param aTarget nsIRDFNode the object of the assertion
 */
ServerObserver.prototype.onAssert = function(aDataSource, aSource, aProperty, aTarget) 
{
    // screen out changes to the top level container
    if (aSource.Value.match(ServerObserver.topLevelRegExp) != null) {
        return;
    }

    var uri = aSource.Value;

    // if this server does not exist, create it
    if (! this.newServers[uri]) {
        this.newServers[uri] = {};
    }

    // set the property  being asserted
    var propertyName = this.removeServerNS(aProperty.Value);
    var propertyValue = aTarget.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

    this.newServers[uri][propertyName] = propertyValue;
    
    // if we've filled in all the necessary values, add the server for real
    if (this.newServers[uri].user && this.newServers[uri].password) {
        // add this server
        this.app.addServer(uri, this.newServers[uri].user, this.newServers[uri].password);

        // if we've added the default post server, set it
        if (this.prefs.getPref("defaultPostServer") == uri) {
            this.prefs.setDefaultPostServer();
        }

        // remove this server from newServers
        delete this.newServers[uri];
    }    
}

/**
 * This method is called whenever an assertion is removed
 * from the data source.
 *
 * @param aDataSource nsIRDFDataSource the datasource that is issuing the notification.
 * @param aSource nsIRDFResource the subject of the assertion
 * @param aProperty nsIRDFResource the predicate of the assertion
 * @param aTarget nsIRDFNode the object of the assertion
 */
ServerObserver.prototype.onUnassert = function(aDataSource, aSource, aProperty, aTarget) {

    // screen out changes to the top level container
    if (aSource.Value.match(ServerObserver.topLevelRegExp) != null) {
        return;
    }
    
    var uri = aSource.Value;

     // remove the server, if we haven't already
    if (this.app.hasServer(uri)) {
        this.app.removeServer(uri);
    }
}
             

/**
 * This method is called when the object of an assertion
 * changes from one value to another.  
 *
 * @param aDataSource nsIRDFDataSource the datasource that is issuing
 *   the notification.
 * @param aSource nsIRDFResource the subject of the assertion
 * @param aProperty nsIRDFResource the predicate of the assertion
 * @param aOldTarget nsIRDFNode the old object of the assertion
 * @param aNewTarget nsIRDFNode the new object of the assertion
 */
ServerObserver.prototype.onChange = function(aDataSource, aSource, aProperty, aOldTarget, aNewTarget) {
     // screen out changes to the top level container
     if (aSource.Value.match(ServerObserver.topLevelRegExp) != null) {
         return;
     }

     var uri = aSource.Value;

     // if this server does not exist, create it
     if (! this.changeServers[uri]) {
         this.changeServers[uri] = {};
    }

    // set the property being changed
    var propertyName = this.removeServerNS(aProperty.Value);
     var propertyValue = aNewTarget.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

    this.changeServers[uri][propertyName] = propertyValue;
    
    // if we've filled in all the necessary values, change the server for real
    if (this.changeServers[uri].user && this.changeServers[uri].password) {
        // remove the old server
        this.app.removeServer(uri);

        // add the new version
        this.app.addServer(uri, this.changeServers[uri].user, this.changeServers[uri].password);

        // if we've added the default post server, set it
        if (this.prefs.getPref("defaultPostServer") == uri) {
            this.prefs.setDefaultPostServer();
        }

        // remove this server from changeServers
        delete this.changeServers[uri];
    }    
}


/**
 * This method is called when the subject of an assertion
 * changes from one value to another.  In our case, this
 * should not occur.
 *
 * @param aDataSource nsIRDFDataSource the datasource that is issuing
 *   the notification.
 * @param aOldSource nsIRDFResource the old subject of the assertion
 * @param aNewSource nsIRDFResourcethe new subject of the assertion
 * @param aProperty nsIRDFResource the predicate of the assertion
 * @param aTarget nsIRDFNode the object of the assertion
 */
ServerObserver.prototype.onMove = function(aDataSource, aOldSource, aNewSource, aProperty, aTarget) {}

/**
 * This method is called when a datasource is about to
 * send several notifications at once. The observer can
 * use this as a cue to optimize its behavior. The observer
 * can expect the datasource to call endUpdateBatch() when
 * the group of notifications has completed.
 *
 * @param aDataSource nsIRDFDataSource the datasource that is going to
 *   be issuing the notifications.
*/
ServerObserver.prototype.beginUpdateBatch = function(aDataSource) {}

/**
 * This method is called when a datasource has completed
 * issuing a notification group.
 *
 * @param aDataSource nsIRDFDataSource the datasource that has finished
 *   issuing a group of notifications
 */
ServerObserver.prototype.endUpdateBatch = function(aDataSource) {}

/**
 * Strips the server namespace from the beginning
 * of the parameter string.
 *
 * @param str String to remove namespace from
 * @return spliced string
 */
ServerObserver.prototype.removeServerNS = function(str) {
    return str.substring(ServerObserver.SERVER_NAMESPACE.length);
}

